#ifndef ENTIER_H
#define ENTIER_H

#include <iosfwd>
using std::istream;
using std::ostream;
#include <iterator>
using std::iterator;
using std::iterator_traits;
using std::random_access_iterator_tag;
#include <algorithm>
#include "relations.h" // CRTP

#pragma warning(disable: 4290)

class entier
   : relation::equivalence<entier>,
     relation::ordonnancement<entier>
{
public:
   //
   // Reprsentation
   //
   typedef int elem_type;
   typedef short size_type;
   static const size_type nbits = sizeof(elem_type)*8;
   //
   // Cas d'exception
   //
   class DivisionParZero {};
   class Debordement {};
   class HorsBornes {};
private:
   elem_type valeur_;
public:
   template <class T>
      class BaseBit
         : relation::equivalence<BaseBit<T> >,
           relation::logique<BaseBit<T> >,
           relation::bit_a_bit<BaseBit<T> > // spcifier le 2e paramtre?
      {
         mutable T *pvaleur_; // donne partage...
         size_type bitpos_;
         //
         // Pointeur vers mthode d'instance (allumer, eteindre); utilis
         // pour des optimisations primitives dans certaines mthodes
         //
         typedef void (BaseBit::*basemodifop) ();
         void allumer ()
            { *pvaleur_ |= (1 << bitpos_); }
         void eteindre ()
            { *pvaleur_ &= (~(1 << bitpos_)); }
         bool est_allume () const
            { return (*pvaleur_ & (1 << bitpos_)) != 0; }
         bool est_eteint () const
            { return !est_allume (); }
      public:
         BaseBit (T *pValeur, const size_type BitPos)
            : pvaleur_ (pValeur), bitpos_ (BitPos)
         {
         }
         //
         // Construction par copie et destruction: Ok tel que gnrs
         //
         // Affectation spcialise pour certains types
         //
         BaseBit & operator= (const bool b)
         {
            static const basemodifop Op[] =
            {
               &BaseBit::eteindre, &BaseBit::allumer
            };
            (this->*Op[b])();
            return *this;
         }
         BaseBit & operator= (const int val)
         {
            static const basemodifop Op[] =
            {
               &BaseBit::eteindre, &BaseBit::allumer
            };
            (this->*Op[val!=0])();
            return *this;
         }
         BaseBit & operator= (const BaseBit &bb)
         {
            static const basemodifop Op[] =
            {
               &BaseBit::eteindre, &BaseBit::allumer
            };
            (this->*Op[bb.est_allume ()])();
            return *this;
         }
         //
         // relationnels (!= obtenus par CRTP)
         //
         bool operator== (const BaseBit &b) const
            { return est_allume () == b.est_allume (); }
         bool operator== (const bool b) const
            { return est_allume () == b; }
         bool operator== (const int val) const
            { return est_allume () == (val != 0); }
         //
         // logique binaire, sur copie. Ceci contrevient aux consignes
         // de Meyers et cie. Est-ce que le jeu en vaut la chandelle?
         // Les versions inverses sont obtenues par CRTP
         //
         bool operator && (const BaseBit &b) const
            { return est_allume () && b.est_allume (); }
         friend bool operator && (const BaseBit &b0, const bool b1)
            { return b0.est_allume () && b1; }
         friend bool operator && (const BaseBit &b, const int val)
            { return b0.est_allume () && val != 0; }
         friend bool operator && (const int val, const BaseBit &b)
            { return b && val; }
         bool operator || (const BaseBit &b) const
            { return est_allume () || b.est_allume (); }
         friend bool operator || (const BaseBit &b0, const bool b1)
            { return b0.est_allume () || b1; }
         friend bool operator || (const BaseBit &b, const int val)
            { return b0.est_allume () || val != 0; }
         //
         // Ngation logique
         //
         bool operator! () const
            { return est_eteint (); }
         //
         // Bit  bit sur copie. Devrait-on retourner des int?
         // Les versions inverses sont obtenues par CRTP
         //
         T operator & (const BaseBit &b) const
            { return (est_allume () && b.est_allume ())? 1 : 0; }
         friend T operator & (const BaseBit &b0, const bool b1)
            { return (b0.est_allume () && b1)? 1 : 0; }
         friend T operator & (const BaseBit &b, const int val)
            { return (b.est_allume () && (val != 0))? 1 : 0; }
         T operator | (const BaseBit &b) const
            { return (est_allume () || b.est_allume ())? 1 : 0; }
         friend T operator | (const BaseBit &b0, const bool b1)
            { return (b0.est_allume () || b1)? 1 : 0; }
         friend T operator | (const BaseBit &b, const int val)
            { return (b.est_allume () || (val != 0))? 1 : 0; }
         T operator ^ (const BaseBit &b) const
            { return (est_allume () != b.est_allume ())? 1 : 0; }
         friend T operator ^ (const BaseBit &b0, const bool b1)
            { return (est_allume () != b)? 1 : 0; }
         friend T operator ^ (const BaseBit &b, const int val)
            { return (est_allume () != (val != 0))? 1 : 0; }
         T operator ~ () const
            { return est_allume ()? 0 : 1; }
         //
         // Bit  bit sur acteur
         //
         BaseBit & operator &= (const BaseBit &b)
         {
            if (!b) eteindre ();
            return *this;
         }
         BaseBit & operator &= (const bool b)
         {
           if (!b) eteindre ();
            return *this;
         }
         BaseBit & operator &= (const int val)
         {
            if (!val) eteindre ();
            return *this;
         }
         BaseBit & operator |= (const BaseBit &b)
         {
            if (b) allumer ();
            return *this;
         }
         BaseBit & operator |= (const bool b)
         {
            if (b) allumer ();
            return *this;
         }
         BaseBit & operator |= (const int val)
         {
            if (val) allumer ();
            return *this;
         }
         //
         // Conversion implicite. Utiliser l'idiome Testable?
         //
         operator bool () const
            { return est_allume (); }
         //
         // Contourner l'asymtrie bool et Bit
         //
         void swap (BaseBit &bb)
         {
            bool etatAutre = bb;
            const basemodifop Op[] =
            {
               &BaseBit::eteindre, &BaseBit::allumer
            };
            (bb.*Op[est_allume()])();
            (this->*Op[etatAutre])();
         }
      };
   //
   // Abstraction du concept de valeur (modifiable)
   //
   typedef BaseBit <elem_type> Bit;
   //
   // Abstraction du concept de valeur (non modifiable)
   //
   typedef BaseBit <const elem_type> ConstBit;

   //
   // Itrateur  accs direct, mouvement vers l'avant
   //
private:
   //
   //
   //
   template <class ElemT, class PosT>
      struct AvanceurTraits
      {
         typedef ElemT elem_type;
         typedef PosT size_type;
         static const size_type PositionDebut = 0;
         static const size_type PositionFin   = sizeof (elem_type) * 8;
      };
   //
   //
   //
   template <class ElemT, class PosT>
      struct ReculeurTraits
      {
         typedef ElemT elem_type;
         typedef PosT size_type;
         static const size_type PositionDebut = sizeof (elem_type) * 8 - 1;
         static const size_type PositionFin   = -1;
      };
   template <class PosT, class TraitsT>
      struct Deplaceur
      {
         typedef PosT offset_type;
         static const offset_type
            DEBUT = TraitsT::PositionDebut,
            FIN = TraitsT::PositionFin;
      };
   template <class T, class PosT, class TraitsT = AvanceurTraits<T, PosT> >
      struct Avanceur
         : Deplaceur<PosT, TraitsT>
      {
         static void avancer (offset_type &pos)
            throw (HorsBornes)
         {
            if (++pos > FIN) throw HorsBornes ();
         }
         static void avancer (offset_type &pos, const size_type n)
            throw (HorsBornes)
               { if (pos += n > FIN) throw HorsBornes (); }
         static void reculer (offset_type &pos)
            throw (HorsBornes)
               { if (--pos < DEBUT) throw HorsBornes (); }
         static void reculer (offset_type &pos, const size_type n)
            { if (pos -= n < DEBUT) throw HorsBornes (); }
         static bool precede (const offset_type &posA, const offset_type &posB)
            { return posA < posB; }
         static bool est_legal (const offset_type pos)
            { return pos >= DEBUT && pos <= FIN; } // tolrance  end()
      };
   template <class T, class PosT, class TraitsT = ReculeurTraits<T, PosT> >
      struct Reculeur
         : Deplaceur<PosT, TraitsT>
      {
         static void avancer (offset_type &pos)
            throw (HorsBornes)
               { if (--pos < FIN) throw HorsBornes (); }
         static void avancer (offset_type &pos, const size_type n)
            throw (HorsBornes)
               { if (pos -= n < FIN) throw HorsBornes (); }
         static void reculer (offset_type &pos)
            throw (HorsBornes)
               { if (++pos > DEBUT) throw HorsBornes (); }
         static void reculer (offset_type &pos, const size_type n)
            { if (pos += n > DEBUT) throw HorsBornes (); }
         static bool precede (const offset_type posA, const offset_type posB)
            { return posB < posA; }
         static bool est_legal (const offset_type pos)
            { return pos >= FIN && pos <= DEBUT; } // tolrance  end()
      };
public:
   template <class T, class Bougeur>
      class bititt
         : relation::equivalence<bititt<T, Bougeur> >,
           relation::ordonnancement<bititt<T, Bougeur> >,
           public std::iterator<random_access_iterator_tag, T, size_type>
      {
      public:
         typedef Bougeur bornes_type;
         typedef T elem_wrap_t;
      private:
         elem_type *pvaleur_;
         size_type bitpos_;
         size_type GetBitPos () const throw()
            { return bitpos_; }
      public:
         bititt (elem_type *pValeur, const size_type BitPos)
            throw (HorsBornes)
            : pvaleur_ (pValeur), bitpos_ (BitPos)
         {
            if (!bornes_type::est_legal (GetBitPos()))
               throw HorsBornes ();
         }
         //
         // Sainte-Trinit: Ok
         //
         bititt & operator++ ()
         {
            bornes_type::avancer(bitpos_);
            return *this;
         }
         bititt operator++ (int)
            throw (HorsBornes)
         {
            bititt b (*this);
            operator++();
            return b;
         }
         bititt & operator-- ()
            throw (HorsBornes)
         {
            bornes_type::reculer(bitpos_);
            return *this;
         }
         bititt operator-- (int)
            throw (HorsBornes)
         {
            bititt b (*this);
            operator--();
            return b;
         }
         //
         // Comparaisons d'itrateurs (!= obtenu par CRTP)
         //
         bool operator== (const bititt &b) const
            throw (HorsBornes)
         {
            if (pvaleur_ != b.pvaleur_) throw HorsBornes ();
            return GetBitPos () == b.GetBitPos ();
         }
         //
         // Relation d'ordre (complt par CRTP)
         //
         bool operator< (const bititt &b) const
            throw (HorsBornes)
         {
            if (pvaleur_ != b.pvaleur_) throw HorsBornes ();
            return bornes_type::precede (GetBitPos (), b.GetBitPos ());
         }
         //
         // Dplacement binaire, sur copie. La soustraction n'est pas commutative
         //
         friend bititt operator+ (const bititt &b, const int n)
            throw (HorsBornes)
               { return bititt (b.pvaleur_, b.GetBitPos () + n); }
         friend bititt operator+ (const int n, const bititt &b)
            throw (HorsBornes)
               { return b + n; }
         bititt operator- (const size_type n) const
            throw (HorsBornes)
               { return bititt (pvaleur_, GetBitPos () - n); }
         //
         // Indirection
         //
         elem_wrap_t operator[] (const size_type n)
            { return *(*this+n); }
         //
         // Dplacement binaire, sur acteur
         //
         bititt& operator+= (const size_type n)
            throw (HorsBornes)
         {
            bornes_type::avancer(bitpos_, n);
            return *this;
         }
         bititt& operator-= (const size_type n)
            throw (HorsBornes)
         {
            bornes_type::reculer(bitpos_, n);
            return *this;
         }
         //
         // Indirection
         //
         elem_wrap_t operator * ()
            throw (HorsBornes)
         {
            if (!pvaleur_) throw HorsBornes ();
            return elem_wrap_t (pvaleur_, GetBitPos ());
         }
         // hum...
         void swap (bititt &_b)
            { (*_b).swap (*(*this)); }
      };

   //
   // Itrateur bit  bit (avant, modifiable)
   //
   typedef bititt<
      Bit, Avanceur <elem_type, size_type>
   > iterator;
   //
   // Itrateur bit  bit (avant, non-modifiable)
   //
   typedef bititt <
      ConstBit, Avanceur <elem_type, size_type>
   > const_iterator;
   //
   // Itrateur bit  bit (arrire, modifiable)
   //
   typedef bititt <
      Bit, Reculeur <elem_type, size_type>
   > reverse_iterator;
   //
   // Itrateur bit  bit (arrire, non modifiable)
   //
   typedef bititt <
      ConstBit, Reculeur <elem_type, size_type>
   > const_reverse_iterator;

   //
   // les versions constantes reposent sur des classes qui garantissent l'intgrit
   //  du elem_type manipul; le const_cast<elem_type*>() est calcul et scuritaire.
   //
   // Les gnrateurs sont utilitaires seulement et allgent le code des mthodes
   // crant des itrateurs sur les extrmits.
   //
private:
   template <class Itt, class T>
      friend Itt make_begin_iterator (T &val)
      {
         typedef typename Itt::bornes_type bornes_type;
         return Itt (&val, bornes_type::DEBUT);
      }
   template <class Itt, class T>
      friend Itt make_begin_iterator (const T &val)
      {
         typedef typename Itt::bornes_type bornes_type;
         return Itt (&const_cast<T&>(val), bornes_type::DEBUT);
      }
   template <class Itt, class T>
      friend Itt make_end_iterator (T &val)
      {
         typedef typename Itt::bornes_type bornes_type;
         return Itt (&val, bornes_type::FIN);
      }
   template <class Itt, class T>
      friend Itt make_end_iterator (const T &val)
      {
         typedef typename Itt::bornes_type bornes_type;
         return Itt (&const_cast<T&>(val), bornes_type::FIN);
      }
public:
   iterator begin ()
      { return make_begin_iterator<iterator>(valeur_); }
   const_iterator begin () const
      { return make_begin_iterator<const_iterator>(valeur_); }
   reverse_iterator rbegin ()
      { return make_begin_iterator<reverse_iterator> (valeur_); }
   const_reverse_iterator rbegin () const
      { return make_begin_iterator<const_reverse_iterator> (valeur_); }
   iterator end ()
      { return make_end_iterator<iterator> (valeur_); }
   const_iterator end () const
      { return make_end_iterator<const_iterator> (valeur_); }
   reverse_iterator rend ()
      { return make_end_iterator<reverse_iterator> (valeur_); }
   const_reverse_iterator rend () const
      { return make_end_iterator<const_reverse_iterator> (valeur_); }

   //
   // Accesseur de premier ordre
   //
   elem_type valeur () const throw()
      { return valeur_; }

   //
   // Construction/ destruction. Les clauses throw() prsument
   // que elem_type est un type entier primitif ou quelque chose
   // d'aussi simple et scuritaire (je pourrais le tester par
   // des assertions statiques)
   //
   entier () throw()
      : valeur_ (0)
   {
   }
   entier (const elem_type Valeur) throw()
      : valeur_ (Valeur)
   {
   }
   entier (const bool Valeur) throw()
      : valeur_ (Valeur? 1 : 0)
   {
   }
   //
   // Sainte-Trinit: Ok
   //
   //
   // Relationnels (!= obtenu par CRTP)
   //
   bool operator== (const entier &e) const throw()
      { return valeur () == e.valeur (); }
   //
   // Ordonnancement (complt par CRTP)
   //
   bool operator< (const entier &e) const throw()
      { return valeur () < e.valeur (); }
   //
   // Arithmtique (binaires sur copies)
   //
   entier operator + (const entier &e) const throw()
      { return entier (valeur () + e.valeur ()); }
   entier operator - (const entier &e) const throw()
      { return entier (valeur () + e.valeur ()); }
   entier operator * (const entier &e) const throw()
      { return entier (valeur () * e.valeur ()); }
   entier operator / (const entier &e) const
      throw (DivisionParZero)
   {
      if (!e.valeur ()) throw DivisionParZero ();
      return entier (valeur () / e.valeur ());
   }
   entier operator % (const entier &e) const
      throw (DivisionParZero)
   {
      if (!e.valeur ()) throw DivisionParZero ();
      return entier (valeur () % e.valeur ());
   }
   //
   // Arithmtique (binaires sur acteur)
   //
   entier& operator += (const entier &e)
   {
      valeur_ += e.valeur ();
      return *this;
   }
   entier& operator -= (const entier &e)
   {
      valeur_ -= e.valeur ();
      return *this;
   }
   entier& operator *= (const entier &e)
   {
      valeur_ *= e.valeur ();
      return *this;
   }
   entier& operator /= (const entier &e)
      throw (DivisionParZero)
   {
      if (!e.valeur ()) throw DivisionParZero ();
      valeur_ /= e.valeur ();
      return *this;
   }
   entier& operator % (const entier &e)
      throw (DivisionParZero)
   {
      if (!e.valeur ()) throw DivisionParZero ();
      valeur_ %= e.valeur ();
      return *this;
   }
   //
   // Arithmtique (unaire)
   //
   entier operator - () const throw()
      { return entier (-valeur_); }
   entier operator + () const throw()
      { return *this; }
   //
   // Autoincrment/ autodcrment
   //
   entier& operator ++ ()
   {
      ++valeur_;
      return *this;
   }
   entier& operator -- ()
   {
      --valeur_;
      return *this;
   }
   entier operator ++ (int)
   {
      entier e(*this);
      operator++();
      return e;
   }
   entier operator -- (int)
   {
      entier e(*this);
      operator--();
      return e;
   }
   //
   // Bit  bit (binaire sur copies)
   //
   entier operator & (const entier &e) const throw()
      { return entier (valeur () & e.valeur ()); }
   entier operator | (const entier &e) const throw()
      { return entier (valeur () | e.valeur ()); }
   entier operator ^ (const entier &e) const throw()
      { return entier (valeur () ^ e.valeur ()); }
   entier operator << (const entier &e) const
      throw (Debordement)
   {
      if (e.valeur () >= nbits) throw Debordement ();
      return entier (valeur () << e.valeur ());
   }
   entier operator >> (const entier &e) const
      throw (Debordement)
   {
      if (e.valeur () >= nbits) throw Debordement ();
      return entier (valeur () >> e.valeur ());
   }
   //
   // Bit  bit (binaire sur acteur)
   //
   entier& operator &= (const entier &e)
   {
      valeur_ &= e.valeur ();
      return *this;
   }
   entier& operator |= (const entier &e)
   {
      valeur_ |= e.valeur ();
      return *this;
   }
   entier& operator ^= (const entier &e)
   {
      valeur_ ^= e.valeur ();
      return *this;
   }
   entier& operator <<= (const entier &e)
      throw (Debordement)
   {
      if (e.valeur () >= nbits) throw Debordement ();
      valeur_ <<= e.valeur ();
      return *this;
   }
   entier& operator >>= (const entier &e)
   {
      if (e.valeur () >= nbits) throw Debordement ();
      valeur_ >>= e.valeur ();
      return *this;
   }
   //
   // Bit  bit (unaire)
   //
   entier operator~ () const throw()
      { return entier (~valeur ()); }
   //
   // Logique (binaire, sur copie). Contrevient aux recommandations
   // de Meyers et cie. Le jeu en vaut-il la chandelle?
   //
   entier operator && (const entier &e) const throw()
      { return entier (valeur () && e.valeur ()); }
   entier operator || (const entier &e) const throw()
      { return entier (valeur () || e.valeur ()); }
   //
   // Logique (unaire, sur acteur)
   //
   bool operator ! () const throw()
      { return !valeur (); }
   //
   // Accs immdiat
   //
   const_iterator::elem_wrap_t
      operator[] (const int n) const throw (HorsBornes)
         { return *(begin()+n); }
   iterator::elem_wrap_t
      operator[] (const int n) throw (HorsBornes)
         { return *(begin()+n); }
};

ostream & operator<< (ostream &, const entier &);
istream & operator>> (istream &, entier &);

namespace std
{
   //
   // Il faut spcialiser std::iter_swap sur des itrateurs non constants
   // d'entier car *itt sur un tel itrateur retourne un Bit, et *b sur un
   // bit retourne un bool (ce qui est asymtrique).
   //
   template <>
      void iter_swap (entier::iterator, entier::iterator);
   template <>
      void iter_swap (entier::reverse_iterator, entier::reverse_iterator);
}

#pragma warning(default: 4290)

#endif